package CGI::PersistentParams;
use strict;
use URI;
use base 'Class::Accessor';
use vars qw($VERSION);
$VERSION = '0.06';

=head1 NAME

CGI::PersistentParams - make query parameters persist between pages

=head1 SYNOPSIS

  # Take query from CGI.pm
  my $q = CGI::PersistentParams->new(
    discard => [qw[action pass user]]
  );

  # Explicitly set a parameter
  $q->param( template => 'foo' );
  $q->param( page => 1 );

  print $q->link( { template => 'bar' } );
  # outputs "template=bar;page=1"

  print $q->link( { page => 2 } );
  # outputs "template=foo;page=2"

=head1 METHODS

=cut

__PACKAGE__->mk_accessors(qw(cgi discard));

=head2 C<< new ARGS >>

Returns a new query parameter handler. The
following arguments can be passed to the constructor:

  cgi     - the CGI query.
            Will be taken from CGI.pm by default.
  discard - The parameters that will not be taken from 
            the cgi object

=cut

sub new {
  my ($class,%args) = @_;

  $args{cgi} ||= do {
    require 'CGI.pm';
    CGI->import();
    CGI->new();
  };

  $args{discard} ||= [];

  my $self = \%args;
  bless $self, $class;
  $self;
};

=head2 C<< link ITEMS, DEFAULTS >>

Returns a new link out of current values
and the passed items. Duplicate items
are not possible.

C<ITEMS> must be a hash reference.

Any items in the result that have the same
string-value as in the C<DEFAULTS> hashref
will be omitted from the resulting string.

Currently does not URI-encode values.

=cut

sub link {
  my ($self,$args, $defaults) = @_;
  $args ||= {};
  $defaults ||= {};
  my $V = $self->merged($args);
  
  for (keys %$defaults) {
      if (exists $V->{$_} and $V->{$_} eq $defaults->{$_}) {
          delete $V->{$_};
      };
  };
  
  return URI->new(CGI->new($V)->url(-full => 1, -query => 1))->query || "";
};

=head2 C<< hidden ITEMS, DEFAULTS >>

Returns a string containing the
defaults and the passed items as
hidden form elements. Duplicate items
are not possible.

C<ITEMS> must be a hash reference.

Currently does not HTML-encode values.

=cut

sub hidden {
  my ($self,$args,$defaults) = @_;
  
  $args ||= {};
  $defaults ||= {};
  
  my $V = $self->merged($args);

  for (keys %$defaults) {
      if (exists $V->{$_} and $V->{$_} eq $defaults->{$_}) {
          delete $V->{$_};
      };
  };
  
  return (join "", map { sprintf qq{<input type="hidden" name="%s" value="%s" />\n}, $_, $V->{$_} } keys %$V) || "";
};

=head2 C<< merged ARGS >>

Merges the hashref passed via C<< ARGS >>
and returns a fresh hashref to the
merged hash.

All keys mentioned in C<< discard >>
are removed.

=cut

sub merged {
  my ($self,$args) = @_;
  my %V = (%{ $self->cgi->Vars });

  delete $V{$_}
    for @{ $self->discard };

  %V = (%V, %$args);

  \%V
};

=head2 C<< param >>

Passes through to the C<param> method of the
C<cgi> object.

=cut

sub param {
  my ($self) = shift;
  $self->cgi->param(@_)
};

1;

=head1 LICENSE

This module may be used under the same terms as Perl itself

=head1 AUTHOR

Max Maischein, E<lt>corion@cpan.orgE<gt>

=head1 SEE ALSO

L<CGI>.pm
