package Filesys::Virtual::FFRIndexed;
use strict;
# use base 'Filesys::Virtual'; # do we need to?
use base 'Class::Accessor';

use FFRIndexed;
use FFRIndexed::Template; # really?
use Carp qw(croak);
use vars qw($VERSION);
$VERSION = '0.06';

=head1 NAME

Filesys::Virtual::FFRIndexed - export an FFRIndexed database as file system

=head1 SYNOPSIS

  my $fs = Filesys::Virtual::FFRIndexed->new();

=cut

for (qw(open_read close_read open_write close_write seek utime chmod rename delete mkdir rmdir)) {
    no strict 'refs';
    *{ $_ } = sub {
        croak __PACKAGE__ . "::$_ is unimplemented, as this is a read-only filesystem.";
    };
};

__PACKAGE__->mk_accessors(qw(root_path index cwd query columns subdirs));

=head2 C<< $fs->new $ARGREF >>

Instantiates a new filesystem instance. 

The following keys in C<$ARGREF> are recognized:

    root_path - a string that will be prepended/stripped before FFRIndexed is queried
                This should enable "mounting" the FFRIndexed tree/graph anywhere in
                an existing L<Filesys::Virtual> tree.
                
                This parameter currently gets ignored though.
    
    index     - a preinitialized L<FFRIndexed> instance. If this is not
                given, a fresh instance will be created.
                
    cwd       - current path
    
    columns   - the columns used for refinements

The remaining arguments
are passed immediately through to the L<FFRIndexed> instance.

=cut

sub new {
    my ($package,$argref) = @_;
    
    my $root_path = delete $argref->{root_path} || "";
    my $index = delete $argref->{index};
    my $cwd = delete $argref->{cwd} || "/"; # relative to root_path, obviously
    my $columns = delete $argref->{columns} || [qw[mime_type artist album track title ]];
    
    my $self = $package->SUPER::new({});
    
    $index ||= FFRIndexed->new(%$argref);
    
    #warn "Root >$root_path<";
    
    $self->columns($columns);
    $self->root_path($root_path);
    $self->index($index);
    $self->cwd($cwd);
    $self->chdir($cwd);
    
    $self
};

=head2 C<< $fs->login >>

Does nothing

=cut

sub login {};

=head2 C<< $fs->cwd >>

Returns the current (query) path.

=head2 C<< $fs->root_path >>

Returns the C<root_path>, the root of this file system.

=head2 C<< $fs->chmod >>

Unimplemented - the file system is readonly

=head2 C<< $fs->modtime $file >>

Returns the time of last modification of C<$file>

Currently implemented as a stub that always returns 0.

=cut

sub modtime {
    0
};

=head2 C<< $fs->size $file >>

Returns the size in bytes of C<$file>

Currently implemented as a stub that always returns 0.

=cut

sub size {
    0
};

=head2 C<< $fs->delete $file >>

Unimplemented - the file system is readonly

=head2 C<< $fs->chdir $dir >>

Refines the query.

=cut

sub chdir {
    my ($self,$dir) = @_;
    
    if ($dir eq '') {
        $dir = $self->cwd();
    };
    
    if ($self->test('d',$dir)) {
        warn "Changing to '$dir'";
        $self->query( $self->requery($dir));
        $self->subdirs({ map { $_->basename => 1 } $self->dirs( $self->query, 1 ) });
        #warn Dumper($self->subdirs);
        return $self->cwd($dir);
    } else {
        warn "Changing to '$dir' failed, it doesn't exist";
        return
    };
};

=head2 C<< $fs->requery $directory >>

Returns the cached C<$query> object if C<$directory>
is equal to C<$cwd>.

=cut

sub requery {
    my ($self,$dir) = @_;
    
    warn sprintf "Requery: '%s' <=> '%s'", $dir, $self->cwd;
    
    if ($dir ne $self->cwd) {
        $self->get_query($dir)
        
    } else {
        return $self->{query} ||= $self->get_query($dir);
    };
};

=head2 C<< $fs->get_query $directory >>

Unconditionally retrieves the L<FFRIndexed> object
for a query path.

=cut

sub get_query {
    my ($self,$dir) = @_;
    use HTTP::Server::FFRIndexed;
    $dir = $self->root_path . $dir;
    warn "Getting query for '$dir'";
    my ($q,$parts) = HTTP::Server::FFRIndexed->decode_query($dir);
    
    return FFRIndexed::Template->new(
        index => $self->index,
        query => $q,
        columns => [@{ $self->columns }],
    );
};

=head2 C<< $fs->folder_contents >>

Returns a list of objects representing the current
contents of the folder and additionally all subfolders.

=cut

sub folder_contents {
    my ($self,$folder) = @_;
    
    return $folder->resultset_list(0,undef), $self->dirs($folder);
};

sub dirs {
    my ($self, $folder, $reread) = @_;
    
    if ($folder eq $self->cwd and not $reread) {
        return keys %{ $self->subdirs }

    } else {
        use Data::Dumper;
        warn sprintf "Reading directories of '%s'", Dumper($folder->query);
        my @dirs;
        for my $col ($folder->refinements) {
            push @dirs, map { Filesys::Virtual::FFRIndexed::Directory->new("$col:$_") } $col->value_list;
            warn Dumper \@dirs;
        };
        @dirs
    };
};

=head2 C<< $fs->mkdir $dir >>

Unimplemented - the file system is readonly

=head2 C<< $fs->rmdir $dir >>

Unimplemented - the file system is readonly

=head2 C<< $fs->list $dir >>

Returns the names of the contents of the directory as a (sorted) list.

If C<$dir> actually is a file, its name is returned.

=cut

sub list {
    my ($self, $dir) = @_;
    
    warn "Listing $dir";
    
    my $folder = $self->requery($dir);
    my @c = $self->folder_contents($folder);
    #warn Dumper \@c;
    return map { $_->basename } @c;
};

=head2 C<< $fs->list_details $dir >>

Returns the names of the contents of the directory as a list
in C<ls -l> format (which is then, subsequently, unpacked again
from the wrapping application ...).

=cut

sub list_details {
    my ($self,$dir) = @_;
    my $folder = $self->requery($dir);
    return map { $self->ls_stat($_) } $self->folder_contents($folder);
};

=head2 C<< $fs->ls_stat $item >>

Returns a string containing C<$item> in C<ls -l> format

=cut

sub ls_stat {
    my ($self,$item) = @_;
    warn Dumper $item;
    my @entries = map { $item->$_ } qw(name);
    return sprintf "%s", @entries;
};

=head2 C<< $fs->stat $file >>

Returns the information as a C<stat> call would

Currently implemented as a list of nine zeroes.

=cut

sub stat {
  0,0,0,0,0,0,0,0,0,0
};

=head2 C<< $fs->test $test, $item >>

Performs the file system test C<$test> on C<$item>

=cut

my %tests = (
    d => \&test_d,
    e => \&test_e,
);

sub test {
    my ($self,$cond,$item) = @_;
    
    my $test = $tests{$cond} || sub {
        warn sprintf "Unimplemented test '%s' called on '%s' (returned false)", $cond, $item;
        return
    };
    
    my $res = $test->($self,$item);
    #warn sprintf "Testing -%s '%s' (%s)", $cond, $item, $res ? 1 : 0;
    
    $res
};

sub test_d {
    my ($self,$item) = @_;
    
    $item =~ m!^(.*/)([^/]*)$!
        or warn "Weirdo itemname '$item'";
    my ($dir,$fn) = ($1,$2);
    
    
    if ($item eq $self->root_path) {
        return 1
    } elsif ($item eq $self->cwd) {
        return 1
    } else {
        if ($dir eq $self->cwd) {
            #use Data::Dumper;
            #warn Dumper( $self->subdirs );
            #warn "-d $fn";
            return exists $self->subdirs->{$fn};
            
        } else {
            warn "(Re)Scanning '$dir' for '$fn'";
            warn sprintf "Cwd is '%s'", $self->cwd ;
            my $folder = $self->requery($dir);
            my @dirs = $self->dirs($folder);
            return scalar grep { $_ eq $fn } @dirs;
        };
    };
};

sub test_e {
    my ($self,$item) = @_;
    
    #if ($item eq $self->cwd) {
        return 1
    #};
};

package Filesys::Virtual::FFRIndexed::Directory;
use strict;

sub new {
    my ($package,$name) = @_;
    bless \$name,$package;
};

sub filename { ${$_[0]} };
sub basename { warn ${$_[0]}; ${$_[0]} };

1;

