package Table::Denormalized;
use strict;

=head1 NAME

Table::Denormalized - Dynamically add columns to a table

=head1 SYNOPSIS

  use Table::Denormalized;
  my $table = Table::Denormalized->new(
    table     => 'my_table',
    primary   => '_id',
    row_class => 'My::Special::Row',
  );

  $table->create;

  # somewhere else, later:

  $table->add_set(@columnsets);
  $table->alter_table;

  # The table now has all columns as specified in the column sets

=cut

use base 'Class::Accessor';
use SQL::Abstract::SuperSearch;
use Carp qw(croak);
use Data::Dumper;
use DBI;
use Scalar::Util 'weaken';
use overload
  '""' => sub { shift->table },
  fallback => 1, # to make Template::Toolkit happy :-(
  ;

use Table::Denormalized::Column;
use Table::Denormalized::ColumnSet;
use Table::Denormalized::Row;
use Table::Denormalized::ResultSet;

use vars qw($VERSION);
$VERSION = '0.06';

__PACKAGE__->mk_accessors(qw(columnsets dbh sql table primary row_class));

=head1 ABSTRACT

This module provides the scaffolding to dynamically
add (and in the future also remove) columns from a
database table.

=head2 C<< new >>

Constructs a new instance.

  dsn               DBI connect string
  dbh               A valid dbh - may be given instead of the DSN
  columnsets 		The column sets, can also be added via add_set
  table             The name of the table, default is 'file_cache'
  primary           The name of the primary key, default is '_id'
  row_class         The class single rows will be blessed into. The default
                    is "$class::Row", which should inherit from
                    Table::Denormalized::Row

=cut

sub new {
  my ($class,%args) = @_;
  if (exists $args{dbh}) {
    # all is well
  } elsif (exists $args{dsn}) {
    $args{dbh} = DBI->connect(delete $args{dsn})
  };

  $args{sql} ||= SQL::Abstract::SuperSearch->new();

  my $columns = delete $args{columns};

  my $self = $class->SUPER::new({
    columnsets => [],
    table     => 'file_cache',
    primary   => '_id',
    row_class => (__PACKAGE__ . "::Row"),
    %args
  });

  #warn $self->row_class;

  if ($columns) {
    $self->add_set( Table::Denormalized::ColumnSet->new( @$columns ));
  };

  # Warn somehow if $self->primary is not in $self->row_class->skeleton ?

  $self
};

=head2 C<< create >>

Issues the C<CREATE TABLE> command.

=cut

sub create {
  my ($self) = @_;
  my $dbh = $self->dbh;
  {
    local $dbh->{PrintError};
    local $dbh->{RaiseError};
    $dbh->do(sprintf 'drop table %s', $self->table);
  };
  $dbh->do(sprintf 'create table %s (%s INTEGER PRIMARY KEY)', $self->table, $self->primary);
};

=head2 C<< columns >>

Enumerates the columns and returns them in alphabetical (C<sort>) order.

=cut

sub columns {
  my $self = shift;
  my %columns;
  for my $set (@{ $self->columnsets }) {
    for my $col (@{ $set->columns }) {
      $columns{$col->name} = $col;
    };
  };
  map {$columns{$_}} sort keys %columns
};

=head2 C<< add_set SETS >>

Adds one or more new column sets to the table, but does not
yet alter the table.

=cut

sub add_set {
  my $self = shift;
  push @{ $self->columnsets }, @_;
  $self;
};

=head2 C<< missing_columns >>

Returns the columns that are in the column sets but are
not yet in the table structure of the database.

=cut

sub missing_columns {
  my $self = shift;
  my (@columns) = $self->columns;
  my @existing;
  my $sth = $self->dbh->prepare(sprintf 'select * from %s where 1 = 0', $self->table);
  $sth->execute();

  my (@missing) = grep {! exists $sth->{NAME_lc_hash}->{ lc $_->name }} ($self->columns);
  $sth->finish;

  return @missing;
};

=head2 C<< alter_sql >>

Returns the SQL commands that need to be executed to
alter the table structure to what the column sets specify.

Currently will only generate C<ADD COLUMN> statements and never
C<DROP COLUMN> statements. This seems to be sensible because
different programs might want to use the same table.

=cut

sub alter_sql {
  my $self = shift;
  my @missing_columns = $self->missing_columns;
  return
    unless @missing_columns;

  map { sprintf "alter table %s add column %s", $self->table, $_->as_sql } @missing_columns;
};

=head2 C<< alter_table >>

Actually executes the statements returned by C<alter_sql>
and thus brings the table in synch with the specification.

=cut

sub alter_table {
  my ($self,$verbose) = @_;
  for my $col ($self->alter_sql) {
    warn $col
      if $verbose;
    $self->dbh->do($col)
      or warn $col;
  };

  $self->dbh->commit
    unless $self->dbh->{AutoCommit};

  1;
};

sub update_or_insert {
  my ($self,$data,$primary) = @_;

  $primary ||= $self->primary;

  croak "Didn't find column $primary in " . Dumper($data)
    unless exists $data->{$primary};

  my $rows;
  if (defined $primary && $data->{$primary}) {
    my %values = ref $data eq 'HASH' ? %$data : $data->body;
    my ($sql,@bind) = $self->sql->update($self->table, \%values,
                                           { $primary => $data->{$primary}});

    my $sth = $self->dbh->prepare_cached($sql);
    $rows = $sth->execute(@bind);
    $sth->finish;
  };

  if ($rows == 0) {
    # Need an insert:
    my ($sql,@bind) = $self->sql->insert($self->table, $data);
    my $sth = $self->dbh->prepare_cached($sql);
    my $rows = $sth->execute(@bind);
    $sth->finish;

    # I should write a test for that condition
    die $sth->errstr
      unless $rows > 0;

    my $id = $self->primary;

    if (! exists $data->{$id}) {
      # Need to fetch the (fresh) id from the database
      # Ah, the joy of restricting myself to SQLite
      $data->{$id} = $self->dbh->last_insert_id(undef,undef,$self->table,$self->primary);

      # This should be rolled up into a transaction that does (INSERT; SELECT last_insert_id)
    };
  };

  my $blessed = $self->rows_to_objects([$data]);
  return $blessed->[0]
};

=head2 C<< rows_to_objects ROWREF >>

Blesses all rows in the array reference given by ROWREF
into the desired class and adds a weak backlink to
the table to each row.

If you want to switch this behaviour to use
a real class constructor or to use a different
class like a C<DBIx::Class> derivative, you will
have to override this method in your subclass
or replace this method directly.

Replacing this method should be fairly sane since you will likely
want to use L<DBIx::Class> in your whole project if you use it in
one place already.

=cut

sub rows_to_objects {
  my ($self,$rows) = @_;
  my $class = $self->row_class;
  my $weak = $self;

  for (@$rows) {
    bless $_, $class;
    $_->{_table} = $weak;
  };
  $rows
};

=head2 C<< query WHERE >>

Returns the blessed results of an SQL::Abstract C<select WHERE>
expression. The results will be stored in an array reference
which is blessed into a Table::Denormalized::Resultset.

=cut

sub query {
    my ($self,$query) = @_;
    $query ||= {};
    $self->query_ex( where => $query, order_by => [$self->primary]);
};

=head2 C<< query_ex ARGS >>

(for lack of a better name) runs a full
pass-through to a SQL::Abstract::SuperSearch query
retrieving all columns unless given.

=cut

# Split up in execute/fetch so we can live
# with large resultsets instead of pulling
# them into memory at once
sub query_ex {
    my ($self,%args) = @_;
    my $table = $self->table;
    my @columns = map {$_->name} $self->columns;
    my $columns = delete $args{columns} || [$self->primary,@columns];
    my ($sql,@bind) = $self->sql->select($table,$columns,%args);
    #warn $sql;
    #warn Dumper \@bind;
    
    my $sth = $self->dbh->prepare_cached($sql);
    $sth->execute(@bind);
    
    my $res = $sth->fetchall_arrayref({});
    my $results = Table::Denormalized::ResultSet->new( $res, $args{where} );
    
    $self->rows_to_objects($results);
    
    $results
};

=head2 C<< retrieve ITEMS >>

Shortcut to retrieve rows by their primary key

=cut

sub retrieve {
  my ($self) = shift;
  $self->query( { $self->primary => { -in => \@_ }} )
};

=head2 C<< all >>

Returns all rows stored in the table.

=cut

sub all { shift->query({}) };

=head2 C<< delete ITEMS >>

Deletes all rows as given by their primary key values.

=cut

# Split up in execute/fetch
sub delete {
  my ($self) = shift;
  my $table = $self->table;
  my ($sql,@bind) = $self->sql->delete($table,{ $self->primary => { -in => \@_ }});
  my $sth = $self->dbh->prepare_cached($sql);
  $sth->execute(@bind);

  # Wipe the cache, if any
};

=head2 C<< update ITEMS >>

Saves the data from the hashrefs given by ITEMS into
the table.

=cut

# Split up in execute/fetch
sub update {
  my ($self) = shift;
  my $table = $self->table;
  for my $row (@_) {
    my ($sql,@bind) = $self->sql->update($table,{ $row->_body },
                                           { $self->primary => $row->{$self->primary} });
    my $sth = $self->dbh->prepare_cached($sql);
    $sth->execute(@bind);
  };

  # Wipe/reload/update the cache, if any
};

1;

=head1 A WORD FROM THE AUTHOR

This module is clearly an example of the "Worse Is Better" idea
- the underlying database "design" is horrible and has many drawbacks,
chiefly that it's not relational, but it has the huge advantage if simplifying
and streamlining the code that operates on the table and with the results.

The module came in to existence after I had failed to implement five generations
of well-designed database schemata, mostly because the handling was too complex
to understand after the code had rotted for a week.

=head1 AUTHOR

Max Maischein, E<lt>corion@cpan.orgE<gt>

=head1 SEE ALSO

L<DBI>, A good book on database normalization, L<Class::DBI>, L<DBIx::Class>
